@@ -410,10 +410,10 @@ impl<'a> GraphQueryParser<'a> {
             subqueries.push(self.has_node_type(node_type, fields.node_type));
         }
 
-        if let Some(ref node_subtype) = node.node_subtype {
-            if !node_subtype.is_empty() {
-                subqueries.push(self.has_node_subtype(node_subtype, fields.node_subtype));
-            }
+        if let Some(ref node_subtype) = node.node_subtype
+            && !node_subtype.is_empty()
+        {
+            subqueries.push(self.has_node_subtype(node_subtype, fields.node_subtype));
         }
 
         subqueries
@@ -425,10 +425,10 @@ impl<'a> GraphQueryParser<'a> {
 
         match expression {
             Expression::Value(Relation { value, relation_type }) => {
-                if let Some(value) = value {
-                    if !value.is_empty() {
-                        subqueries.push((Occur::Must, self.has_relation_label(value)));
-                    }
+                if let Some(value) = value
+                    && !value.is_empty()
+                {
+                    subqueries.push((Occur::Must, self.has_relation_label(value)));
                 };
 
                 relation_type.map(|relation_type| {
@@ -437,10 +437,10 @@ impl<'a> GraphQueryParser<'a> {
             }
 
             Expression::Not(Relation { value, relation_type }) => {
-                if let Some(value) = value {
-                    if !value.is_empty() {
-                        subqueries.push((Occur::MustNot, self.has_relation_label(value)));
-                    }
+                if let Some(value) = value
+                    && !value.is_empty()
+                {
+                    subqueries.push((Occur::MustNot, self.has_relation_label(value)));
                 };
 
                 relation_type.map(|relation_type| {
