@@ -26,6 +26,7 @@ use crate::data_store::{DataStore, DataStoreV1, DataStoreV2, OpenReason, Paragra
 use crate::formula::Formula;
 use crate::hnsw::*;
 use crate::inverted_index::{FilterBitSet, InvertedIndexes, build_indexes};
+use crate::vector_types::rabitq;
 use crate::{ParagraphAddr, VectorAddr, VectorErr, VectorR, VectorSegmentMeta, VectorSegmentMetadata};
 use core::f32;
 use io::{BufWriter, Write};
@@ -38,7 +39,7 @@ use std::iter::empty;
 use std::path::Path;
 
 /// How much expensive is to find a node via HNSW compared to a simple brute force scan
-const HNSW_COST_FACTOR: usize = 200;
+const HNSW_COST_FACTOR: usize = 100;
 
 mod file_names {
     pub const HNSW: &str = "index.hnsw";
@@ -71,6 +72,7 @@ pub fn open(metadata: VectorSegmentMetadata, config: &VectorConfig) -> VectorR<O
     #[cfg(not(target_os = "windows"))]
     {
         index.advise(memmap2::Advice::Random)?;
+        index.advise(memmap2::Advice::WillNeed)?;
     }
 
     let inverted_indexes = InvertedIndexes::open(path, metadata.records)?;
@@ -120,7 +122,7 @@ pub fn merge(segment_path: &Path, operants: &[&OpenSegment], config: &VectorConf
             .iter()
             .map(|dp| (dp.alive_paragraphs(), dp.data_store.as_ref()))
             .collect();
-        DataStoreV2::merge(segment_path, node_producers, &config.vector_type)?;
+        DataStoreV2::merge(segment_path, node_producers, config)?;
         let data_store = DataStoreV2::open(segment_path, &config.vector_type, OpenReason::Create)?;
         merge_indexes(segment_path, data_store, operants, config)
     }
@@ -133,7 +135,8 @@ fn merge_indexes<DS: DataStore + 'static>(
     config: &VectorConfig,
 ) -> VectorR<OpenSegment> {
     // Check if the first segment has deletions. If it doesn't, we can reuse its HNSW index
-    let has_deletions = operants[0].alive_paragraphs().count() < operants[0].data_store.stored_paragraph_count();
+    let has_deletions =
+        operants[0].alive_paragraphs().count() < operants[0].data_store.stored_paragraph_count() as usize;
 
     let mut index;
     let start_vector_index;
@@ -149,10 +152,10 @@ fn merge_indexes<DS: DataStore + 'static>(
     let merged_vectors_count = data_store.stored_vector_count();
 
     // Creating the hnsw for the new node store.
-    let retriever = Retriever::new(&[], &data_store, config, -1.0);
+    let retriever = Retriever::new(&data_store, config, -1.0);
     let mut ops = HnswOps::new(&retriever, false);
     for id in start_vector_index..merged_vectors_count {
-        ops.insert(Address(id), &mut index);
+        ops.insert(VectorAddr(id), &mut index);
     }
 
     let hnsw_path = segment_path.join(file_names::HNSW);
@@ -179,14 +182,14 @@ fn merge_indexes<DS: DataStore + 'static>(
 
     let metadata = VectorSegmentMetadata {
         path: segment_path.to_path_buf(),
-        records: data_store.stored_paragraph_count(),
+        records: data_store.stored_paragraph_count() as usize,
         index_metadata: VectorSegmentMeta {
             tags: operants[0].metadata.index_metadata.tags.clone(),
         },
     };
 
     build_indexes(segment_path, &data_store)?;
-    let inverted_indexes = InvertedIndexes::open(segment_path, merged_vectors_count)?;
+    let inverted_indexes = InvertedIndexes::open(segment_path, merged_vectors_count as usize)?;
     let alive_bitset = FilterBitSet::new(metadata.records, true);
 
     Ok(OpenSegment {
@@ -230,7 +233,7 @@ pub fn create(path: &Path, elems: Vec<Elem>, config: &VectorConfig, tags: HashSe
             tags,
         )
     } else {
-        DataStoreV2::create(path, elems, &config.vector_type)?;
+        DataStoreV2::create(path, elems, config)?;
         create_indexes(
             path,
             DataStoreV2::open(path, &config.vector_type, OpenReason::Create)?,
@@ -256,10 +259,10 @@ fn create_indexes<DS: DataStore + 'static>(
 
     // Creating the HNSW using the mmaped nodes
     let mut index = RAMHnsw::new();
-    let retriever = Retriever::new(&[], &data_store, config, -1.0);
+    let retriever = Retriever::new(&data_store, config, -1.0);
     let mut ops = HnswOps::new(&retriever, false);
     for id in 0..vector_count {
-        ops.insert(Address(id), &mut index)
+        ops.insert(VectorAddr(id), &mut index)
     }
 
     {
@@ -281,12 +284,12 @@ fn create_indexes<DS: DataStore + 'static>(
 
     let metadata = VectorSegmentMetadata {
         path: path.to_path_buf(),
-        records: data_store.stored_paragraph_count(),
+        records: data_store.stored_paragraph_count() as usize,
         index_metadata: VectorSegmentMeta { tags },
     };
 
     build_indexes(path, &data_store)?;
-    let inverted_indexes = InvertedIndexes::open(path, vector_count)?;
+    let inverted_indexes = InvertedIndexes::open(path, vector_count as usize)?;
     let alive_bitset = FilterBitSet::new(metadata.records, true);
 
     Ok(OpenSegment {
@@ -300,49 +303,63 @@ fn create_indexes<DS: DataStore + 'static>(
 
 pub struct Retriever<'a, DS: DataStore> {
     similarity_function: fn(&[u8], &[u8]) -> f32,
-    vector_count: usize,
-    temp: &'a [u8],
     data_store: &'a DS,
     min_score: f32,
 }
 impl<'a, DS: DataStore> Retriever<'a, DS> {
-    pub fn new(temp: &'a [u8], data_store: &'a DS, config: &VectorConfig, min_score: f32) -> Retriever<'a, DS> {
+    pub fn new(data_store: &'a DS, config: &VectorConfig, min_score: f32) -> Retriever<'a, DS> {
         Retriever {
-            temp,
             data_store,
             similarity_function: config.similarity_function(),
-            vector_count: data_store.stored_vector_count(),
             min_score,
         }
     }
-    fn find_vector(&'a self, x: Address) -> VectorRef<'a> {
-        self.data_store.get_vector(x.into())
-    }
 }
 
 impl<DS: DataStore> DataRetriever for Retriever<'_, DS> {
-    fn will_need(&self, x: Address) {
-        self.data_store.will_need(x.into());
+    fn will_need(&self, x: VectorAddr) {
+        self.data_store.will_need(x);
     }
 
-    fn get_vector(&self, x @ Address(addr): Address) -> &[u8] {
-        if addr == self.vector_count {
-            self.temp
+    fn will_need_vector(&self, x: VectorAddr) {
+        if self.data_store.has_quantized() {
+            self.data_store.will_need_quantized(x);
         } else {
-            self.find_vector(x).vector()
+            self.data_store.will_need(x);
         }
     }
-    fn similarity(&self, x @ Address(a0): Address, y @ Address(a1): Address) -> f32 {
-        if a0 == self.vector_count {
-            let y = self.find_vector(y).vector();
-            (self.similarity_function)(self.temp, y)
-        } else if a1 == self.vector_count {
-            let x = self.find_vector(x).vector();
-            (self.similarity_function)(self.temp, x)
-        } else {
-            let x = self.find_vector(x).vector();
-            let y = self.find_vector(y).vector();
-            (self.similarity_function)(x, y)
+
+    fn get_vector(&self, x: VectorAddr) -> &[u8] {
+        self.data_store.get_vector(x).vector()
+    }
+
+    fn similarity(&self, x: VectorAddr, y: &SearchVector) -> f32 {
+        match y {
+            SearchVector::Stored(vector_addr) => {
+                let x = self.data_store.get_vector(x).vector();
+                let y = self.data_store.get_vector(*vector_addr).vector();
+                (self.similarity_function)(x, y)
+            }
+            SearchVector::Query(query) => {
+                let x = self.data_store.get_vector(x).vector();
+                (self.similarity_function)(x, query)
+            }
+            SearchVector::RabitQ(query) => {
+                let x = self.data_store.get_quantized_vector(x);
+                let (est, _err) = query.similarity(x);
+                est
+            }
+        }
+    }
+
+    fn similarity_upper_bound(&self, x: VectorAddr, y: &SearchVector) -> EstimatedScore {
+        match y {
+            SearchVector::RabitQ(query) => {
+                let x = self.data_store.get_quantized_vector(x);
+                let (est, err) = query.similarity(x);
+                EstimatedScore::new_with_error(est, err)
+            }
+            _ => EstimatedScore::new_exact(self.similarity(x, y)),
         }
     }
 
@@ -350,8 +367,8 @@ impl<DS: DataStore> DataRetriever for Retriever<'_, DS> {
         self.min_score
     }
 
-    fn paragraph(&self, x: Address) -> ParagraphAddr {
-        self.find_vector(x).paragraph()
+    fn paragraph(&self, x: VectorAddr) -> ParagraphAddr {
+        self.data_store.get_vector(x).paragraph()
     }
 }
 
@@ -394,8 +411,8 @@ pub struct ScoredVector<'a> {
 }
 
 impl ScoredVector<'_> {
-    fn new(addr: Address, data_store: &dyn DataStore, score: f32) -> ScoredVector {
-        let node = data_store.get_vector(addr.into());
+    fn new(addr: VectorAddr, data_store: &dyn DataStore, score: f32) -> ScoredVector<'_> {
+        let node = data_store.get_vector(addr);
         ScoredVector { score, vector: node }
     }
     pub fn score(&self) -> f32 {
@@ -448,11 +465,11 @@ impl OpenSegment {
         self.data_store.size_bytes() + self.index.len() + self.inverted_indexes.space_usage()
     }
 
-    pub fn get_paragraph(&self, id: ParagraphAddr) -> ParagraphRef {
+    pub fn get_paragraph(&self, id: ParagraphAddr) -> ParagraphRef<'_> {
         self.data_store.get_paragraph(id)
     }
 
-    pub fn get_vector(&self, id: VectorAddr) -> VectorRef {
+    pub fn get_vector(&self, id: VectorAddr) -> VectorRef<'_> {
         self.data_store.get_vector(id)
     }
 
@@ -464,7 +481,7 @@ impl OpenSegment {
         results: usize,
         config: &VectorConfig,
         min_score: f32,
-    ) -> Box<dyn Iterator<Item = ScoredVector> + '_> {
+    ) -> Box<dyn Iterator<Item = ScoredVector<'_>> + '_> {
         if let Some(v1) = self.data_store.as_any().downcast_ref::<DataStoreV1>() {
             self._search(v1, query, filter, with_duplicates, results, config, min_score)
         } else if let Some(v2) = self.data_store.as_any().downcast_ref::<DataStoreV2>() {
@@ -475,19 +492,25 @@ impl OpenSegment {
     }
 
     #[allow(clippy::too_many_arguments)]
-    fn _search<DS: DataStore>(
-        &self,
-        data_store: &DS,
+    fn _search<'a, DS: DataStore>(
+        &'a self,
+        data_store: &'a DS,
         query: &[f32],
         filter: &Formula,
         with_duplicates: bool,
         results: usize,
         config: &VectorConfig,
         min_score: f32,
-    ) -> Box<dyn Iterator<Item = ScoredVector> + '_> {
-        let encoded_query = config.vector_type.encode(query);
-        let retriever = Retriever::new(&encoded_query, data_store, config, min_score);
-        let query_address = Address(self.metadata.records);
+    ) -> Box<dyn Iterator<Item = ScoredVector<'a>> + '_> {
+        let raw_query = SearchVector::Query(config.vector_type.encode(query));
+        let rabitq = data_store.has_quantized() && !config.flags.contains(&flags::DISABLE_RABITQ_SEARCH.to_string());
+        let encoded_query = if rabitq {
+            let rabitq = rabitq::QueryVector::from_vector(query, &config.vector_type);
+            &SearchVector::RabitQ(rabitq)
+        } else {
+            &raw_query
+        };
+        let retriever = Retriever::new(data_store, config, min_score);
 
         let mut filter_bitset = self.inverted_indexes.filter(filter);
         if let Some(ref mut bitset) = filter_bitset {
@@ -503,42 +526,63 @@ impl OpenSegment {
         let expected_traversal_scan = results * self.metadata.records / count;
 
         if count < expected_traversal_scan * HNSW_COST_FACTOR {
-            let mut scored_results = Vec::new();
-            for paragraph_addr in bitset.iter() {
-                let paragraph = data_store.get_paragraph(paragraph_addr);
-
-                // Only return the best vector match per paragraph
-                let best_vector_score = paragraph
-                    .vectors(&paragraph_addr)
-                    .map(|va| {
-                        let address = va.into();
-                        let score = retriever.similarity(query_address, address);
-                        Cnx(address, score)
-                    })
-                    .max_by(|v, w| v.1.total_cmp(&w.1))
-                    .unwrap();
-
-                if best_vector_score.1 >= min_score {
-                    scored_results.push(Reverse(best_vector_score));
-                }
-            }
-            scored_results.sort();
-            return Box::new(
-                scored_results
+            self.brute_force_search(bitset, results, retriever, encoded_query, &raw_query)
+        } else {
+            let ops = HnswOps::new(&retriever, true);
+            let neighbours = ops.search(encoded_query, self.index.as_ref(), results, bitset, with_duplicates);
+            Box::new(
+                neighbours
                     .into_iter()
-                    .map(|Reverse(a)| ScoredVector::new(a.0, self.data_store.as_ref(), a.1))
+                    .map(|(address, dist)| ScoredVector::new(address, self.data_store.as_ref(), dist))
                     .take(results),
-            );
+            )
+        }
+    }
+
+    fn brute_force_search<'a, DS: DataStore>(
+        &'a self,
+        bitset: &FilterBitSet,
+        results: usize,
+        retriever: Retriever<'a, DS>,
+        encoded_query: &SearchVector,
+        raw_query: &SearchVector,
+    ) -> Box<dyn Iterator<Item = ScoredVector<'a>> + '_> {
+        let mut scored_results = Vec::new();
+        for paragraph_addr in bitset.iter() {
+            let paragraph = retriever.data_store.get_paragraph(paragraph_addr);
+
+            // Only return the best vector match per paragraph (only relevant with multi-vectors)
+            let best_vector_score = paragraph
+                .vectors(&paragraph_addr)
+                .map(|va| {
+                    let score = retriever.similarity_upper_bound(va, encoded_query);
+                    (va, score)
+                })
+                .max_by(|v, w| v.1.score.total_cmp(&w.1.score))
+                .unwrap();
+
+            if best_vector_score.1.upper_bound >= retriever.min_score {
+                scored_results.push(best_vector_score);
+            }
         }
 
-        let ops = HnswOps::new(&retriever, true);
-        let neighbours = ops.search(query_address, self.index.as_ref(), results, bitset, with_duplicates);
-        Box::new(
-            neighbours
-                .into_iter()
-                .map(|(address, dist)| (ScoredVector::new(address, self.data_store.as_ref(), dist)))
-                .take(results),
-        )
+        if matches!(encoded_query, SearchVector::RabitQ(_)) {
+            // If using RabitQ, rerank top results using the raw vectors
+            Box::new(
+                rabitq::rerank_top(scored_results, results, &retriever, raw_query)
+                    .into_iter()
+                    .map(|Reverse(Cnx(addr, score))| ScoredVector::new(addr, self.data_store.as_ref(), score)),
+            )
+        } else {
+            // If using raw vectors, sort by score and take top_k
+            scored_results.sort_unstable_by(|a, b| b.1.score.total_cmp(&a.1.score));
+            Box::new(
+                scored_results
+                    .into_iter()
+                    .map(|a| ScoredVector::new(a.0, self.data_store.as_ref(), a.1.score))
+                    .take(results),
+            )
+        }
     }
 }
 
@@ -560,7 +604,7 @@ mod test {
     use super::{Elem, create, merge};
     use nidx_protos::prost::*;
 
-    const DIMENSION: usize = 128;
+    const DIMENSION: usize = 256;
 
     fn random_vector(rng: &mut impl Rng) -> Vec<f32> {
         let v: Vec<f32> = (0..DIMENSION).map(|_| rng.gen_range(-1.0..1.0)).collect();
@@ -776,7 +820,7 @@ mod test {
         )?;
 
         // Search a few times
-        let correct = (0..100)
+        let correct: f32 = (0..100)
             .map(|_| {
                 // Search near an existing segment (simulates that the query is related to the data)
                 let base_v = elems.values().nth(rng.gen_range(0..elems.len())).unwrap();
@@ -791,16 +835,22 @@ mod test {
                     .iter()
                     .map(|r| dp.data_store.get_paragraph(r.paragraph()).id().to_string())
                     .collect();
-                let brute_force: Vec<_> = similarities.iter().take(5).map(|r| r.0.clone()).collect();
-                search == brute_force
-            })
-            .filter(|x| *x)
-            .count();
 
-        let recall = correct as f32 / 100.0;
+                let brute_force: Vec<_> = similarities.iter().take(5).map(|r| r.0.clone()).collect();
+                let mut r = 0.0;
+                for b in brute_force {
+                    if search.contains(&b) {
+                        r += 0.2;
+                    }
+                }
+                r
+            })
+            .sum();
+
+        let recall = correct / 100.0;
         println!("Assessed recall = {recall}");
-        // Expected 0.90-0.92, has a little margin because HNSW can be non-deterministic
-        assert!(recall >= 0.88);
+        // Expected ~0.98, has a little margin because HNSW can be non-deterministic
+        assert!(recall >= 0.95);
 
         Ok(())
     }
